import typing

from starlette.routing import Route as StarletteRoute

from .. import Container


class Route(StarletteRoute):
    def __init__(
        self,
        container: Container,
        path: str,
        endpoint: typing.Callable,
        *,
        methods: typing.List[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ):
        wrapped_endpoint = container.partial(endpoint)
        super().__init__(
            path,
            wrapped_endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )


class StarletteContainer(Container):
    def route(
        self,
        path: str,
        endpoint: typing.Callable,
        *,
        methods: typing.List[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ) -> Route:
        return Route(
            self,
            path,
            endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )



from typing import TypeVar

INTROSPECTION_ATTRS = (
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
)

T = TypeVar("T")


def decorated_wrapper(wrapped: T, original) -> T:
    for attr in INTROSPECTION_ATTRS:
        if hasattr(original, attr):
            setattr(wrapped, attr, getattr(original, attr))
    return wrapped



from flask import Flask

from lagom import Container


class FlaskContainer(Container):
    flask_app: Flask

    def __init__(self, app: Flask, container=None):
        self.flask_app = app
        super().__init__(container)

    def route(self, rule, **options):
        def decorator(f):
            endpoint = options.pop("endpoint", None)
            injected_func = self.partial(f)
            self.flask_app.add_url_rule(rule, endpoint, injected_func, **options)
            return f

        return decorator



from starlette.routing import Route

from lagom import Container
from lagom.integrations.starlette import StarletteContainer


class MyDep:
    pass


class ComplexDep:
    def __init__(self, something):
        pass


def some_handler(request, dep: MyDep):
    return "ok"


def test_a_special_starlette_container_can_be_used_and_provides_routes():
    sc = StarletteContainer()
    route = sc.route("/", some_handler)
    assert isinstance(route, Route)
    assert route.endpoint({}) == "ok"


def test_the_starlette_container_can_wrap_an_existing_container(container: Container):
    container[ComplexDep] = ComplexDep("")
    sc = StarletteContainer(container)
    assert isinstance(sc[ComplexDep], ComplexDep)
